package com.outr.webcommunicator.netty

import handler.RequestHandler
import org.jboss.netty.handler.codec.http.HttpRequest
import org.powerscala.event.Listenable
import org.powerscala.property.{ListProperty, Property}
import org.jboss.netty.channel.{MessageEvent, ChannelHandlerContext}
import com.outr.webcommunicator.URL

/**
 * @author Matt Hicks <matt@outr.com>
 */
class MutableWebResource(implicit val webapp: Webapp) extends Listenable with WebResource {
  def parent = webapp
  def name = () => getClass.getSimpleName

  val matchers = new Property[List[HttpRequest => Boolean]](default = Some(Nil)) with ListProperty[HttpRequest => Boolean] {
    def apply(f: => Boolean) = {
      this += ((request: HttpRequest) => f)
    }
  }
  /**
   * Iterated over before loader or creator is called to allow pre-handling if necessary.
   */
  val preChecks = new Property[List[HttpRequest => RequestResult]](default = Some(Nil)) with ListProperty[HttpRequest => RequestResult] {
    def apply(f: => RequestResult) = {
      this += ((request: HttpRequest) => f)
    }
  }
  /**
   * Called if no preChecks are defined to load a cached RequestHandler if one exists.
   */
  val loader = new Property[HttpRequest => RequestHandler]() {
    def apply(f: => RequestHandler) = {
      value = (request: HttpRequest) => f
    }
  }
  /**
   * Called when no preChecks or loader returns and a new instance must be created.
   */
  val creator = new Property[HttpRequest => RequestHandler]() {
    def apply(f: => RequestHandler) = {
      value = (request: HttpRequest) => f
    }
  }

  def regex(regex: String) = (request: HttpRequest) => request.path.matches(regex)
  def matches(path: String) = (request: HttpRequest) => request.path == path
  def urlRegex(regex: String) = (request: HttpRequest) => request2URL(request).toString
  def urlMatcher(matcher: URL => Boolean) = (request: HttpRequest) => matcher(request2URL(request))

  def request(webapp: NettyWebapp, context: ChannelHandlerContext, event: MessageEvent) = event.getMessage match {
    case request: HttpRequest if (matchers().find(m => m(request)).nonEmpty) => {
      apply(context, request)
    }
    case _ => None
  }

  protected def apply(context: ChannelHandlerContext, request: HttpRequest) = {
    val pre = preChecks().view.map(f => f(request)).collectFirst {
      case rr if (rr != RequestResult.Continue) => rr
    }
    pre match {
      case Some(rr) => rr match {
        case handle: RequestResult.Handle => Some(handle.handler)
        case RequestResult.Bypass => None
      }
      case None => {
        val loaded = if (loader() != null) {
          loader()(request)
        } else {
          null
        }
        if (loaded != null) {
          Some(loaded)
        } else {
          Some(creator()(request))
        }
      }
    }
  }
}