package com.outr.webcommunicator.netty

import handler.{PathHandler, RequestHandler}
import org.jboss.netty.bootstrap.ServerBootstrap
import org.jboss.netty.channel.socket.nio.NioServerSocketChannelFactory
import java.util.concurrent.{ConcurrentHashMap, Executors}
import java.net.{URL => JURL, InetSocketAddress}
import org.jboss.netty.channel._
import org.jboss.netty.handler.codec.http._
import org.jboss.netty.handler.stream.ChunkedWriteHandler
import scala.Some
import org.jboss.netty.buffer.ChannelBuffers
import org.jboss.netty.util.CharsetUtil
import annotation.tailrec
import java.util.concurrent.atomic.AtomicBoolean
import org.powerscala.log.Logging
import com.outr.webcommunicator.{URL, Method}
import java.nio.channels.ClosedChannelException
import websocketx.WebSocketHandshakeException

/**
 * @author Matt Hicks <mhicks@powerscala.org>
 */
trait NettyWebapp extends Webapp with ChannelPipelineFactory with Logging {
  NettyWebapp.instance = this

  private var channels = Map.empty[String, Channel]
  private var _resources = List.empty[WebResource]
  def resources = _resources

  val bootstrap = new ServerBootstrap(new NioServerSocketChannelFactory(Executors.newCachedThreadPool(), Executors.newCachedThreadPool()))
  bootstrap.setPipelineFactory(this)

  private val initialized = new AtomicBoolean(false)

  /**
   * Creates a URL from the given values. This method allows overriding the internally seen URL.
   */
  def createURL(method: Method,
                protocol: String,
                host: String,
                port: Int,
                path: String,
                parameters: Map[String, List[String]]) = {
    URL(method, protocol, host, port, path, parameters)
  }

  def bind(host: String, port: Int) = synchronized {
    val address = host match {
      case null => new InetSocketAddress(port)
      case _ => new InetSocketAddress(host, port)
    }
    channels += "%s:%s".format(host, port) -> bootstrap.bind(address)
    if (initialized.compareAndSet(false, true)) {
      initialize()
    }
  }

  def unbind(host: String, port: Int) = synchronized {
    val key = "%s:%s".format(host, port)
    channels.get(key) match {
      case Some(channel) => {
        channel.unbind()
        channels -= key
      }
      case None =>
    }
  }

  def shutdown() = synchronized {
    channels.foreach {
      case (key, channel) => {
        channel.unbind()
        channels -= key
      }
    }
  }

  def register(resource: WebResource): Unit = if (!resources.contains(resource)) synchronized {
    _resources = (resource :: resources).sortBy(h => h.priority.value)
  }

  def register(path: String, resource: String): Unit = register(URLResource(path, resource))

  def register(path: String, url: JURL): Unit = register(URLResource(path, url))

  def register(resource: String): Unit = register(URLResource(if (resource.startsWith("/")) {
    resource
  } else {
    "/%s".format(resource)
  }, resource))

  def registerPath(webPath: String, internalPath: String, cacheContent: Boolean = true) = {
    register(PathHandler(webPath, internalPath, cacheContent = cacheContent))
  }

  def getPipeline = {
    val pipeline = Channels.pipeline()

    pipeline.addLast("decoder", new HttpRequestDecoder)
    pipeline.addLast("aggregator", new HttpChunkAggregator(Int.MaxValue))
    pipeline.addLast("encoder", new HttpResponseEncoder)
    pipeline.addLast("deflater", new HttpContentCompressor)
    pipeline.addLast("streamer", new ChunkedWriteHandler)
    pipeline.addLast("handler", new ConnectionHandler)

    pipeline
  }

  protected def notFound(context: ChannelHandlerContext, event: MessageEvent) = {
    val response = RequestHandler.createResponse(status = HttpResponseStatus.NOT_FOUND, contentType = "text/plain; charset=UTF-8")
    response.setContent(ChannelBuffers.copiedBuffer(response.getStatus.toString, CharsetUtil.UTF_8))
    context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  protected def forbidden(context: ChannelHandlerContext, event: MessageEvent) = {
    val response = RequestHandler.createResponse(status = HttpResponseStatus.FORBIDDEN, contentType = "text/plain; charset=UTF-8")
    response.setContent(ChannelBuffers.copiedBuffer(response.getStatus.toString, CharsetUtil.UTF_8))
    context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  protected def error(context: ChannelHandlerContext, event: MessageEvent, t: Throwable) = {
    t.printStackTrace()
    val response = RequestHandler.createResponse(status = HttpResponseStatus.INTERNAL_SERVER_ERROR, contentType = "text/plain; charset=UTF-8")
    response.setContent(ChannelBuffers.copiedBuffer(response.getStatus.toString, CharsetUtil.UTF_8))
    context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  @tailrec
  private def lookupHandler(context: ChannelHandlerContext, event: MessageEvent, resources: List[WebResource]): Option[RequestHandler] = {
    if (resources.isEmpty) {
      None
    } else {
      val resource = resources.head
      resource.request(this, context, event) match {
        case Some(handler) => Some(handler)
        case None => lookupHandler(context, event, resources.tail)
      }
    }
  }

  protected def contextualize[T](context: ChannelHandlerContext, event: MessageEvent)(f: => T): T = {
    f
  }

  protected def invokeHandler(context: ChannelHandlerContext, event: MessageEvent, handler: RequestHandler) = {
    handler(this, context, event)
  }

  protected def exceptionCaught(context: ChannelHandlerContext, t: Throwable) = {
    val handler = connections.get(context.getChannel)
    if (t.getMessage == "An existing connection was forcibly closed by the remote host") {
      warn(t.getMessage)    // This happens when the browser is closed, so we can generally ignore it
    } else if (t.getMessage == "An established connection was aborted by the software in your host machine") {
      warn(t.getMessage)     // This happens when the connection is terminated abruptly on page change
    } else if (t.getMessage == "Connection reset by peer") {
      warn(t.getMessage)     // Connections sometimes drop
    } else if (t.getMessage == "Broken pipe") {
      warn(t.getMessage)     // Connections sometimes
    } else if (t.isInstanceOf[WebSocketHandshakeException]) {
      warn(t.getMessage)     // Bad attempt to handshake
    } else if (t.isInstanceOf[ClosedChannelException]) {
      trace("Channel closed abruptly")
    } else if (handler == null) {
      errorThrown(t)
    } else {
      handler.errorThrown(t)
    }
  }

  protected def errorThrown(t: Throwable) = {
    error("ConnectionHandler.exceptionCaught without handler", t)
  }

  private val connections = new ConcurrentHashMap[Channel, RequestHandler]()

  def messageReceived(context: ChannelHandlerContext, event: MessageEvent) = {
    contextualize(context, event) {
      val channel = event.getChannel
      try {
        val handlerOption = channel.synchronized {
          connections.get(channel) match {
            case null => {
//              val started = System.currentTimeMillis()
              try {
                val ho = lookupHandler(context, event, resources)
                if (ho.nonEmpty) {
                  connections.put(channel, ho.get)
                }
                ho
              } finally {
//                val took = System.currentTimeMillis() - started
//                if (took > 200) {
//                  val request = event.getMessage.asInstanceOf[HttpRequest]
//                  warn(s"${NettyWebapp.this.getClass}.lookupHandler took $took milliseconds: URI: ${request.getUri}.")
//                }
              }
            }
            case handler => Some(handler)
          }
        }
        handlerOption match {
          case Some(handler) => invokeHandler(context, event, handler)
          case None => notFound(context, event)
        }
      } catch {
        case t: Throwable => NettyWebapp.this.exceptionCaught(context, t)
      }
    }
  }

  def connectionClosed(context: ChannelHandlerContext) = {
    connections.remove(context.getChannel)
  }

  class ConnectionHandler extends SimpleChannelUpstreamHandler {
    override def messageReceived(context: ChannelHandlerContext, event: MessageEvent) = if (event.getMessage != null) {
      NettyWebapp.this.messageReceived(context, event)
    }

    override def exceptionCaught(context: ChannelHandlerContext, e: ExceptionEvent) {
      NettyWebapp.this.exceptionCaught(context, e.getCause)
    }

    override def channelClosed(context: ChannelHandlerContext, e: ChannelStateEvent) {
      super.channelClosed(context, e)

      connectionClosed(context)
    }
  }
}

object NettyWebapp {
  private var instance: NettyWebapp = null

  def apply() = instance

  /**
   * Creates a URL from the given values. This method allows overriding the internally seen URL.
   */
  def createURL(method: Method,
                protocol: String,
                host: String,
                port: Int,
                path: String,
                parameters: Map[String, List[String]]) = {
    if (instance != null) {
      instance.createURL(method, protocol, host, port, path, parameters)
    } else {
      URL(method, protocol, host, port, path, parameters)
    }
  }
}